#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif

#include <sys/socket.h>
#include <sys/select.h>
#include <sys/types.h>
#include <sys/time.h>

#include "tunnelit-common.h"
#include "tunnelit-helper.h"
#include "ethertun.h"
#include "net.h"

typedef enum {
	TUNNELIT_OPMODE_CONNECT = 0,
	TUNNELIT_OPMODE_LISTEN = 1,
	TUNNELIT_OPMODE_TCP = 0,
	TUNNELIT_OPMODE_UDP = 2
} opmode_t;

typedef ssize_t (*recvfunc_t)(int, void *, size_t, int);
typedef ssize_t (*sendfunc_t)(int, const void *, size_t, int);

static ssize_t recv_exact(recvfunc_t func, int fd, void *buf_p, size_t len) {
	ssize_t recv_ret;
	ssize_t retval = 0;
	char *buf = buf_p;

	while (len) {
		recv_ret = func(fd, buf, len, 0);
		if (recv_ret <= 0) {
			return(-1);
		}

		retval += recv_ret;
		buf += recv_ret;
		len -= recv_ret;
	}

	return(retval);
}

static ssize_t send_exact(sendfunc_t func, int fd, const void *buf, size_t len) {
	return(recv_exact((recvfunc_t) func, fd, (void *) buf, len));
}

static fd_set *copy_sock_to_tun(int srcsock, int tun, int socks[TUNNELIT_MAX_CHILDREN], int highestchildidx) {
	unsigned char pktlen_bytes[2];
	static fd_set retval;
	static char buf[16384];
	ssize_t recv_ret, send_ret;
	int sock;
	int idx;
	int pktlen;

	FD_ZERO(&retval);

	recv_ret = recv_exact((recvfunc_t) recv, srcsock, pktlen_bytes, sizeof(pktlen_bytes));
	pktlen = (pktlen_bytes[0] << 8) | pktlen_bytes[1];

	if (pktlen > sizeof(buf)) {
		pktlen = sizeof(buf);
	}

	recv_ret = recv_exact((recvfunc_t) recv, srcsock, buf, pktlen);
	if (recv_ret <= 0) {
		DEBUG_LOG("debug", "Got EOF from Socket (%i), closing it.", srcsock);
		FD_SET(srcsock, &retval);
		return(&retval);
	}

	DEBUG_LOG("debug", "Writing packet to TUN device (%i), len = %i", tun, (int) recv_ret);
	send_ret = send_exact((sendfunc_t) write, tun, buf, recv_ret);
	if (send_ret <= 0) {
		DEBUG_LOG("debug", "Got EOF from TUN device (%i), odd.", tun);
		return(&retval);
	}

	for (idx = 0; idx < highestchildidx; idx++) {
		sock = socks[idx];

		if (sock < 0) {
			continue;
		}

		if (sock == srcsock) {
			continue;
		}

		DEBUG_LOG("debug", "[%i] Writing packet to Socket (%i), len = %i", idx, sock, (int) recv_ret);
		send_ret = send_exact((sendfunc_t) send, sock, pktlen_bytes, sizeof(pktlen_bytes));
		send_ret = send_exact((sendfunc_t) send, sock, buf, recv_ret);
		if (send_ret <= 0) {
			DEBUG_LOG("debug", "Got EOF from Socket (%i), closing it.", sock);
			FD_SET(sock, &retval);
			continue;
		}
	}

	return(&retval);
}

static fd_set *copy_tun_to_socks(int tun, int socks[TUNNELIT_MAX_CHILDREN], int highestchildidx) {
	unsigned char pktlen_bytes[2];
	static fd_set retval;
	static char buf[16384];
	ssize_t read_ret, send_ret;
	int sock;
	int idx;

	FD_ZERO(&retval);

	read_ret = read(tun, buf, sizeof(buf));
	if (read_ret <= 0) {
		DEBUG_LOG("debug", "Got EOF from TUN (%i) !?", tun);
		return(&retval);
	}

	pktlen_bytes[0] = (read_ret >> 8) & 0xff;
	pktlen_bytes[1] = read_ret & 0xff;

	for (idx = 0; idx < highestchildidx; idx++) {
		sock = socks[idx];

		if (sock < 0) {
			continue;
		}

		DEBUG_LOG("debug", "[%i] Writing packet to Socket (%i), len = %i", idx, sock, (int) read_ret);
		send_ret = send_exact((sendfunc_t) send, sock, pktlen_bytes, sizeof(pktlen_bytes));
		send_ret = send_exact((sendfunc_t) send, sock, buf, read_ret);
		if (send_ret <= 0) {
			DEBUG_LOG("debug", "Got EOF from Socket (%i), closing it.", sock);
			FD_SET(sock, &retval);
			continue;
		}
	}

	return(&retval);
}

static int net_connect_tcp_persistent(const char *host, int port) {
	int newfd;

	newfd = -1;
	while (newfd < 0) {
		newfd = net_connect_tcp(host, port);
		if (newfd >= 0) {
			break;
		}

		PERROR("net_connect_tcp");
		sleep(5);
	}

	return(newfd);
}

static void print_help(FILE *chan) {
	fprintf(chan, "Usage: tunnelit-p2p [-Vhluo] [-p port] [-H host] [-d devname]\n");
	fprintf(chan, "   -V            Print version and exit\n");
	fprintf(chan, "   -h            Print this help and exit\n");
	fprintf(chan, "   -l            Listen (default is connect)\n");
	fprintf(chan, "   -u            UDP mode (not implemented)\n");
	fprintf(chan, "   -o            Only allow one client to connect at a time.\n");
	fprintf(chan, "   -H <host>     Connect to host, required if \"-l\" not specified\n");
	fprintf(chan, "   -d <devname>  TUN/TAP device name\n");
	return;
}

int main(int argc, char **argv) {
	opmode_t opmode = TUNNELIT_OPMODE_CONNECT | TUNNELIT_OPMODE_TCP;
	fd_set readfds, errfds, *copyfailfds;
	char *host = NULL, *devname = NULL;
	char ch;
	int tunfd, masterfd, newfd, childfds[TUNNELIT_MAX_CHILDREN], maxfd;
	int port = 656;
	int opt_onlyoneclient = 0;
	int found_place_for_fd;
	int select_ret;
	int idx, idxinner;

	while ((ch = getopt(argc, argv, "Vhluop:H:d:")) != -1) {
		switch (ch) {
			case 'V':
				printf("TunnelIt Point-to-Point version %s-%s", PACKAGE_VERSION,
#ifdef DEBUG
				       "debug"
#else
				       "rel"
#endif
				);
				return(0);
				break;
			case 'h':
				print_help(stdout);
				return(0);
				break;
			case 'l':
				opmode |= TUNNELIT_OPMODE_LISTEN;
				break;	
			case 'u':
				opmode |= TUNNELIT_OPMODE_UDP;
				break;
			case 'o':
				opt_onlyoneclient = 1;
				break;
			case 'p':
				port = atoi(optarg);
				break;
			case 'H':
				host = optarg;
				break;
			case 'd':
				devname = strdup(optarg);
				break;
			case '?':
			case ':':
				print_help(stderr);
				return(1);
				break;
		}
	}

	if ((opmode & TUNNELIT_OPMODE_LISTEN) == TUNNELIT_OPMODE_CONNECT) {
		if (!host) {
			ERROR_LOG("Must specify hostname (-H)");
			return(1);
		}
	}

	tunfd = tun_alloc(devname);
	if (tunfd < 0) {
		PERROR("tun_alloc");
		ERROR_LOG("Unable to create tun/tap device");
		return(1);
	}

	for (idx = 0; idx < (sizeof(childfds) / sizeof(childfds[0])); idx++) {
		childfds[idx] = -1;
	}

	masterfd = -1;
	switch ((int) opmode) {
		case TUNNELIT_OPMODE_CONNECT | TUNNELIT_OPMODE_TCP:
			newfd = net_connect_tcp_persistent(host, port);

			childfds[0] = newfd;

			break;
		case TUNNELIT_OPMODE_CONNECT | TUNNELIT_OPMODE_UDP:
			ERROR_LOG("not implemented");
			return(1);
			break;
		case TUNNELIT_OPMODE_LISTEN | TUNNELIT_OPMODE_UDP:
			ERROR_LOG("not implemented");
			return(1);
			break;
		case TUNNELIT_OPMODE_LISTEN | TUNNELIT_OPMODE_TCP:
			masterfd = -1;
			while (masterfd < 0) {
				masterfd = net_listen_tcp(port);
				if (masterfd >= 0) {
					break;
				}

				PERROR("net_listen_tcp");
				sleep(5);
			}

			break;
	}

	while (1) {
		FD_ZERO(&readfds);
		FD_ZERO(&errfds);

		maxfd = 0;
		for (idx = 0; idx < (sizeof(childfds) / sizeof(childfds[0])); idx++) {
			if (childfds[idx] < 0) {
				continue;
			}

			FD_SET(childfds[idx], &readfds);
			FD_SET(childfds[idx], &errfds);
			if (childfds[idx] > maxfd) {
				maxfd = childfds[idx];
			}
		}
		if (tunfd > maxfd) {
			maxfd = tunfd;
		}
		FD_SET(tunfd, &readfds);
		FD_SET(tunfd, &errfds);

		if (masterfd >= 0) {
			if (masterfd > maxfd) {
				maxfd = masterfd;
			}
			FD_SET(masterfd, &readfds);
			FD_SET(masterfd, &errfds);
		}

		select_ret = select(maxfd + 1, &readfds, NULL, &errfds, NULL);
		if (select_ret < 0) {
			sleep(5);
			continue;
		}

		if (select_ret) {
			if (FD_ISSET(tunfd, &errfds)) {
				select_ret--;
				ERROR_LOG("TUN device (%i) had an error, aborting.", tunfd);
				return(1);
			}

			if (FD_ISSET(tunfd, &readfds)) {
				select_ret--;
				DEBUG_LOG("debug", "Copying from TUN device (%i) to Sockets...", tunfd);
				copyfailfds = copy_tun_to_socks(tunfd, childfds, TUNNELIT_MAX_CHILDREN);
				for (idx = 0; idx < (sizeof(childfds) / sizeof(childfds[0])); idx++) {
					if (childfds[idx] < 0) {
						continue;
					}

					if (FD_ISSET(childfds[idx], copyfailfds)) {
						DEBUG_LOG("debug", "Child Socket (%i) had an error, dropping it", childfds[idx]);
						close(childfds[idx]);
						childfds[idx] = -1;

						if (idx == 0 && opmode == (TUNNELIT_OPMODE_CONNECT | TUNNELIT_OPMODE_TCP)) {
							sleep(5);
							DEBUG_LOG("debug", "Attempting to reconnect to %s:%i...", host, port);
							childfds[0] = net_connect_tcp_persistent(host, port);
						}
					}
				}
			}
		}

		if (select_ret && masterfd >= 0) {
			if (FD_ISSET(masterfd, &errfds)) {
				select_ret--;
				ERROR_LOG("Listening Socket (%i) had an error, aborting.", masterfd);
				return(1);
			}

			if (FD_ISSET(masterfd, &readfds)) {
				select_ret--;
				DEBUG_LOG("debug", "Accepting new connection (%i)", masterfd);
				newfd = -1;
				while (newfd < 0) {
					newfd = accept(masterfd, NULL, NULL);
					if (newfd >= 0) {
						break;
					}

					PERROR("accept");
					sleep(5);
				}

				if (opt_onlyoneclient) {
					for (idx = 0; idx < (sizeof(childfds) / sizeof(childfds[0])); idx++) {
						if (childfds[idx] >= 0) {
							close(newfd);
							newfd = -1;
						}
					}
				}

				if (newfd >= 0) {
					found_place_for_fd = 0;
					for (idx = 0; idx < (sizeof(childfds) / sizeof(childfds[0])); idx++) {
						if (childfds[idx] < 0) {
							found_place_for_fd = 1;
							childfds[idx] = newfd;
							break;
						}
					}

					if (!found_place_for_fd) {
						close(newfd);
					}
				}
			}
		}

		idx = 0;
		while (select_ret) {
			for (; idx < (sizeof(childfds) / sizeof(childfds[0])); idx++) {
				if (childfds[idx] < 0) {
					continue;
				}

				if (FD_ISSET(childfds[idx], &errfds)) {
					select_ret--;
					DEBUG_LOG("debug", "Child Socket (%i) had an error, dropping it", childfds[idx]);
					close(childfds[idx]);
					childfds[idx] = -1;

					if (idx == 0 && opmode == (TUNNELIT_OPMODE_CONNECT | TUNNELIT_OPMODE_TCP)) {
						sleep(5);
						DEBUG_LOG("debug", "Attempting to reconnect to %s:%i...", host, port);
						childfds[0] = net_connect_tcp_persistent(host, port);
					}
					break;
				}

				if (FD_ISSET(childfds[idx], &readfds)) {
					select_ret--;
					DEBUG_LOG("debug", "Copying from Socket (%i) to TUN device (%i)", childfds[idx], tunfd);
					copyfailfds = copy_sock_to_tun(childfds[idx], tunfd, childfds, TUNNELIT_MAX_CHILDREN);

					for (idxinner = 0; idxinner < (sizeof(childfds) / sizeof(childfds[0])); idxinner++) {
						if (childfds[idxinner] < 0) {
							continue;
						}

						if (FD_ISSET(childfds[idxinner], copyfailfds)) {
							DEBUG_LOG("debug", "Child Socket (%i) had an error, dropping it", childfds[idxinner]);
							close(childfds[idxinner]);
							childfds[idxinner] = -1;

							if (idxinner == 0 && opmode == (TUNNELIT_OPMODE_CONNECT | TUNNELIT_OPMODE_TCP)) {
								sleep(5);
								DEBUG_LOG("debug", "Attempting to reconnect to %s:%i...", host, port);
								childfds[0] = net_connect_tcp_persistent(host, port);
							}
						}
					}

					break;
				}
			}
		}
	}

	close(tunfd);

	return(0);
}
